
import * as Actions from "./actions";
import { IActionData, Type } from "./actions/interfaces";
import { IActionFunctionData } from "./functions/interfaces";
import { ActionFunction } from "./functions/ActionFunction";
import { minimizeXML } from "./utils/minimize-xml";


export interface IBuilderJSON {
    actions: IActionData[];
    functions?: IActionFunctionData[];
    name?: string;
    comments?: string;
    debugger?: any;
}

/**
 * Converts JSON Actions to Shoutpoint CCML.
 */
export class CCMLBuilder {

    private _lastCallbackResponseName = "";

    private _actionConstructors: { [key: string]: any } = {
        CCML            : this.buildCCML,
        COLLECT         : this.buildCollect,
        CONFERENCE      : this.buildConference,
        DETECT          : this.buildDetect,
        DIAL            : this.buildDial,
        FUNCTION        : this.buildFunction,
        GOOGLEANALYTICS : this.buildGoogleAnalytics,
        GOTO            : this.buildGoto,
        HANGUP          : this.buildHangUp,
        IF              : this.buildIf,
        LABEL           : this.buildLabel,
        PAUSE           : this.buildPause,
        PLAY            : this.buildPlay,
        QUEUE           : this.buildQueue,
        RECORD          : this.buildRecord,
        REDIRECT        : this.buildRedirect,
        REJECT          : this.buildReject,
        SAY             : this.buildSay,
        SESSIONDATA     : this.buildSessionData,
        SMS             : this.buildSMS,
        STOP            : this.buildStop,
        TRANSFER        : this.buildTransfer,
        VAR             : this.buildVar,
        WEBHOOK         : this.buildWebhook,
    };

    private _functions: Array<ActionFunction> = [];
    private _actions: Array<Actions.IAction> = [];

    /**
     * CCMLBuilder constructor.
     * @param json - The JSON Actions to parse
     */
    constructor (private json: IBuilderJSON, private _debugger?: Actions.Debugger) {

        if (json.debugger === true && this._debugger === undefined) {
            this._debugger = new Actions.Debugger();
        }

        this._lastCallbackResponseName = json.name || "";

        this.addActions(...json.actions);

        if (Array.isArray(json.functions)) {
            this.addFunctions(...json.functions);
        }
    }

    public getJSONActionsWithIDs() {
        return this.json;
    }

    /**
     * Returns the parsed Actions.
     */
    public getActions(): Array<Actions.IAction> {
        if (this._debugger !== undefined) {
            const db = this._debugger;
            return this._actions.reduce((p, c) => p.concat({ type: Type.Debugger, toCCML: db.toCCML.bind(this._debugger, c) } as any, c), [] as Actions.IAction[]);
        } else {
            return this._actions.slice(0);
        }
    }

    /**
     * Add the JSON Actions to parse. Action array order is preserved.
     * @param data - JSON Actions to parse.
     */
    public addActions(...data: IActionData[]): CCMLBuilder {

        data.forEach(d => {
            // If no d.type, handle special case.
            if (!d.type) {
                // Assignment var special case, convert to Actions.Var...
                Object.keys(d)
                    .filter(v => v.trim()[0] === "$")
                    .forEach(v => this._actions.push(new Actions.Var(v, (d as any)[v])));
            } else {
                // Else, handle regular Action.
                const action = this.callActionConstructor(d);
                this._actions.push(action);
                d.id = action.id; // Add id to JSON for debugger.
            }
        });

        return this;
    }

    /**
     * Returns the parsed JSON Functions.
     */
    public getFunctions (): Array<ActionFunction> {
        return this._functions.slice(0);
    }

    /**
     * Add the JSON Functions to parse. Order is not important.
     * @param funcs - JSON Functions to parse.
     */
    public addFunctions (...funcs: IActionFunctionData[]): CCMLBuilder {
        funcs.forEach(f => this._functions.push(new ActionFunction(f)));
        return this;
    }

    /**
     * Outputs the given JSON Actions to CCML.
     * @param [env="dev"] - Set which environment this CCML will be deployed.
     * @param [isModule=false] - If true, <DOCTYPE> and <CCML> wrappers omitted.
     */
    public build (env: "dev" | "qa" | "prod" = "dev", isModule = false): string {

        let ccml = [];

        if (!isModule) {
            ccml.push(`<!DOCTYPE CCML SYSTEM "http://${env}.shoutpoint.com/TelWeb/ccml.dtd">`);
            ccml.push(`<CCML>`);
            ccml.push(`<SET varName="collect_keys" value="" export="true"/>`);
        }

        if (this._debugger !== undefined) {
            this._debugger.jsonActions = this.getJSONActionsWithIDs();
            // Add begin callback loop debugger.
            ccml.push(this._debugger.toCCML({ id: "", name: "", type: Type.Debugger } as any));
            ccml.push("<SAY>Debugger On. Callback start.</SAY>");
        }

        ccml = ccml.concat(
            this._actions.map(a => {

                let type = Actions.Type[a.type].toUpperCase(),
                    comments = JSON.stringify((a.comments || "").replace(/<[^>]+>/g, ""));

                if (this._debugger !== undefined) {
                    return `${this._debugger.toCCML(a)}\n<!-- ${type} ${comments} -->\n${a.toCCML()}\n`;
                }
                return `\n<!-- ${type} ${comments} -->\n${a.toCCML()}\n`;
            })
        );

        ccml = ccml.concat(this._functions.map(f => f.toCCML())); // TODO Add Debugger

        if (this._debugger !== undefined) {
            // Add end callback loop debugger.
            ccml.push("<SAY>Debugger. Callback end.</SAY>");
        }

        if (!isModule) {
            ccml.push("</CCML>");
        }

        return env !== "dev" ? minimizeXML(ccml.join("")) : ccml.join("");
    }



    /////////////////////////////////////////////////
    // HELPERS


    private eventCCMLBuilder (actions: IActionData[], name = "event") {
        return new CCMLBuilder({ actions, name }, this._debugger);
    }

    private callActionConstructor(d: Actions.IActionData): Actions.IAction {

        d.params = d.params || {};
        return this._actionConstructors[d.type.toUpperCase()].call(this, d);
    }

    private buildCCML(data: Actions.IActionData): Actions.CCML {

        let action = new Actions.CCML(data.params.ccml, data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildCollect(data: Actions.IActionData): Actions.Collect {

        let action = new Actions.Collect(data.name);
        action.comments = data.comments || "";
        action.numDigits = data.params.num_digits || 1;
        action.finishOnKey = data.params.finish_on_key || "pound";
        action.digitsOnly = !!data.params.digitsOnly;
        action.timeout = data.params.timeout || 5000;
        action.actions = this.eventCCMLBuilder(data.actions || [], `${action.name}_event`).getActions();
        return action;
    }

    private buildConference(data: IActionData): Actions.Conference {

        let action = new Actions.Conference(data.params.name, data.name);
        action.comments = data.comments || "";
        action.event = ((evt: string) => {
            switch (evt) {
                case "leave":
                    return Actions.ConferenceEvent.LEAVE;
                case "mute":
                    return Actions.ConferenceEvent.MUTE;
                case "unmute":
                    return Actions.ConferenceEvent.UNMUTE;
                case "hold":
                    return Actions.ConferenceEvent.HOLD;
                case "resume":
                    return Actions.ConferenceEvent.RESUME;
                default:
                    return Actions.ConferenceEvent.JOIN;
            }
        })(data.params.event || "join");
        action.autoHold = String(data.params.auto_hold).toLowerCase() === "true";
        action.joinMuted = String(data.params.join_muted).toLowerCase() === "true";
        action.muteOnKey = String(data.params.mute_on_key || "off").toLowerCase();
        action.unmuteOnKey = String(data.params.unmute_on_key || "off").toLowerCase();
        action.exitOnKey = String(data.params.exit_on_key || "off").toLowerCase();
        action.async = String(data.params.async).toLowerCase() === "true";
        if (typeof data.params.max_participants === "number") {
            action.maxParticipants = Math.max(1, data.params.max_participants);
        }
        action.announceOnJoin = data.params.announce_on_join === "beep" ? "beep" : "none";
        action.holdAudio = data.params.hold_audio || [];
        action.endOnExit = String(data.params.end_on_exit).toLowerCase() === "true";
        action.holdOnExit = String(data.params.hold_on_exit).toLowerCase() === "true";
        action.startOnJoin = String(data.params.start_on_join).toLowerCase() !== "false";
        action.holdOnJoin = String(data.params.hold_on_join).toLowerCase() === "true";

        if (data.events) {
            Object.keys(Actions.ConferenceOnEvent)
                .map(k => Actions.ConferenceOnEvent[k])
                .forEach(e => {
                    action.setEvent(e, this.eventCCMLBuilder(data.events[e] || [], `${action.name}_${e}_event`).getActions());
                });
        }

        return action;
    }

    private buildDial(data: IActionData): Actions.Dial {

        let action = new Actions.Dial(data.params.no, data.name);
        action.comments = data.comments || "";
        action.callerIdNo = data.params.caller_id_no || "${apiNo}";
        action.detection = data.params.detection || "none";
        action.inboundAPI = data.params.inbound_api || "CallFlows";
        action.disableRedialBlock = data.params.disable_redial_block === true;

        if (data.session_data !== undefined) {
            action.sessionData = data.session_data;
        }

        if (data.events) {
            Object.keys(Actions.DialOnEvent)
                .map(k => Actions.DialOnEvent[k])
                .forEach(e => {
                    action.setEvent(e, this.eventCCMLBuilder(data.events[e] || [], `${action.name}_${e}_event`).getActions());
                });
        }

        return action;
    }

    private buildDetect(data: IActionData): Actions.Detect {

        let action = new Actions.Detect(data.name);
        action.comments = data.comments || "";
        action.triggerLiveInMS = data.params.trigger_live_in_ms || 3000;

        if (data.events) {
            Object.keys(Actions.DetectOnEvent)
                .map(k => Actions.DetectOnEvent[k])
                .forEach(e => action.setEvent(e, this.eventCCMLBuilder(data.events[e] || [], `${action.name}_${e}_event`).getActions()));
        }

        return action;
    }

    private buildFunction(data: IActionData): Actions.ActionFunction {

        let action = new Actions.ActionFunction(data.name);
        action.comments = data.comments || "";
        action.args = data.params;
        return action;
    }

    private buildGoogleAnalytics(data: IActionData & { callback_response: any; event: any; }): Actions.GoogleAnalytics {

        let action = new Actions.GoogleAnalytics(data.callback_response, data.params, this._lastCallbackResponseName || data.name);
        action.comments = data.comments || "";
        if (data.event) {
            action.setGAEvent(data.event.action, data.event.label, data.event.category, data.event.value);
        }
        return action;
    }

    private buildGoto(data: IActionData): Actions.Goto {

        let action = new Actions.Goto(data.params.label, data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildHangUp(data: IActionData): Actions.HangUp {

        let action = new Actions.HangUp(data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildIf (data: IActionData): Actions.If {

        let action = new Actions.If(data.condition, this.eventCCMLBuilder(data.then || [], `${data.name}_if_event`).getActions(), data.name);
        action.comments = data.comments || "";
        if (Array.isArray(data.orElse)) {
            action.setEvent(Actions.IfEvent.OnElse, this.eventCCMLBuilder(data.orElse as any || [], `${action.name}_else_event`).getActions());
        }
        return action;
    }

    private buildLabel(data: IActionData): Actions.Label {

        let action = new Actions.Label(data.name || "");
        action.comments = data.comments || "";
        return action;
    }

    private buildPause(data: IActionData): Actions.Pause {

        let action = new Actions.Pause(data.params.seconds, data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildPlay(data: IActionData): Actions.Play {

        let action = new Actions.Play(data.params.url || data.params.dtmf, data.name);
        action.comments = data.comments || "";
        if (!!data.params.url) {
            action.stopOnKey = String(data.params.stop_on_key).toLowerCase() === "true";
        } else {
            action.dtmfDelay = data.params.dtmf_delay || 250;
            action.dtmfDashDelay = data.params.dtmf_dash_delay || 250;
        }
        action.repeat = data.params.repeat || 1;
        action.parseVars = data.params.parse_vars !== false;
        return action;
    }

    private buildQueue(data: IActionData): Actions.Queue {

        let action = new Actions.Queue(data.params.name, data.name);
        action.comments = data.comments || "";
        action.event = ((evt: string) => {
            switch (evt) {
                case "poll":
                    return Actions.QueueEvent.POLL;
                case "poll_connect":
                    return Actions.QueueEvent.POLL_CONNECT;
                case "remove":
                    return Actions.QueueEvent.REMOVE;
                default:
                    return Actions.QueueEvent.OFFER;
            }
        })((data.params.event || "offer").toLowerCase());
        action.async = String(data.params.async).toLowerCase() === "true";
        action.maxCalls = typeof data.params.max_calls === "number" ? +data.params.max_calls : undefined;
        action.holdAudio = data.params.hold_audio !== undefined ? data.params.hold_audio : true;

        if (data.events) {
            Object.keys(Actions.QueueOnEvent)
                .map(k => Actions.QueueOnEvent[k])
                .forEach(e => {
                    action.setEvent(e, this.eventCCMLBuilder(data.events[e] || [], `${action.name}_${e}_event`).getActions());
                });
        }

        return action;
    }

    private buildRecord(data: IActionData): Actions.Record {

        let action = new Actions.Record(data.params.action, data.params.group, data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildRedirect(data: IActionData): Actions.Redirect {

        let action = new Actions.Redirect(data.params.url, data.name);
        action.comments = data.comments || "";
        action.method = data.params.method;
        action.contentType = data.params.content_type;
        action.headers = data.params.headers;
        return action;
    }

    private buildReject(data: IActionData): Actions.Reject {

        let action = new Actions.Reject(data.params && data.params.reason, data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildSay(data: IActionData): Actions.Say {

        let action = new Actions.Say(data.params.text, data.name);
        action.comments = data.comments || "";
        action.repeat = data.params.repeat || 1;
        action.stopOnKey = String(data.params.stop_on_key).toLowerCase() === "true";
        action.gender = data.params.gender || "default";
        action.language = data.params.language || "default";
        action.voice = data.params.voice || "default";
        action.parseVars = data.params.parse_vars !== false;
        return action;
    }

    private buildSessionData(data: IActionData): Actions.SessionData {

        let action = new Actions.SessionData(data.session_data, data.name);
        action.comments = data.comments || "";
        return action;
    }

    private buildSMS(data: IActionData): Actions.SMS {

        let action = new Actions.SMS(data.params.message, data.name);
        action.no = data.params.no || "${callerNo}";
        action.callerIdNo = data.params.caller_id_no || "${apiNo}";
        action.comments = data.comments || "";
        return action;
    }

    private buildStop(data: IActionData): Actions.Stop {

        let action = new Actions.Stop(data.name);
        action.actions = (data.params.type || "AUDIO").toUpperCase();
        action.comments = data.comments || "";
        return action;
    }

    private buildTransfer(data: IActionData): Actions.Transfer {

        let action = new Actions.Transfer(data.params.no, data.name);
        action.comments = data.comments || "";
        action.event = data.params.event || "start";
        action.callerIdNo = data.params.caller_id_no || "${callerNo}";
        action.async = String(data.params.async).toLowerCase() === "true";
        return action;
    }

    private buildVar(data: IActionData): Actions.Var {

        let action = new Actions.Var(data.params.assignmentVar, data.params.expression, data.name);
        if (data.comments) { action.comments = data.comments; }
        return action;
    }

    private buildWebhook(data: IActionData): Actions.Webhook {

        let action = new Actions.Webhook(data.params.event_name, data.name);
        action.comments = data.comments || "";
        if (data.params.url) { action.url = String(data.params.url); }
        if (data.params.content_type) { action.contentType = data.params.content_type; }
        if (data.params.headers) { action.headers = data.params.headers; }
        if (data.params.method) { action.method = data.params.method; }
        action.awaitResponse = !!data.params.await_response;
        action.webhookData = data.params.webhook_data || {};

        if (data.events) {
            Object.keys(Actions.WebhookOnEvent)
                .map(k => Actions.WebhookOnEvent[k])
                .forEach(e => {
                    action.setEvent(e, this.eventCCMLBuilder(data.events[e] || [], `${action.name}_${e}_event`).getActions());
                });
        }

        return action;
    }
}
