import {
    IActionData,
    CoreAction,
    Type,
} from "./interfaces";
import {VarParser} from "../utils";

export enum ConferenceEvent {
    JOIN, LEAVE, MUTE, UNMUTE, HOLD, RESUME
}

export const ConferenceOnEvent: any = {
    OnEnter                     : "on_enter",
    OnEnterAlone                : "on_enter_alone",
    OnEnterStarted              : "on_enter_started",
    OnEnterWaiting              : "on_enter_waiting",
    OnEnterError                : "on_enter_error",
    OnEnterErrorMaxParticipants : "on_enter_error_max_participants",
    OnEnterErrorEnded           : "on_enter_error_ended"
};

/**
     {
         "type": "CONFERENCE",

        "params": {
            "name": "my-conference",
            "event": "join",
            "mute_on_key": "1",
            "unmute_on_key": "3",
            "exit_on_key": "#",

            "async": false,

            "max_participants": 20,

            "hold_audio": ["http://file.wav"],

            "join_muted": false,

            "hold_on_join": false,
            "start_on_join": false,

            "end_on_exit": false,
            "hold_on_exit": false
        },

        "events": {
            "on_enter_started": [],
            "on_enter_waiting": [],
            "on_enter_alone": [],
            "on_enter": [],

            "on_enter_error_max_participants": [],
            "on_enter_error_ended": [],
            "on_enter_error": []
        }
    }
*/
export class Conference extends CoreAction {

    private _paramMap: any = {
        id               : "id",
        event            : "control",
        auto_hold        : "holdAuto",
        join_muted       : "joinMuted",
        mute_on_key      : "mute",
        unmute_on_key    : "unmute",
        exit_on_key      : "leave",
        announce         : "announce",

        max_participants : "maxParticipants",
        async            : "async",             // boolean

        $onJoin          : "onJoin",            // hold, start
        $onLeave         : "onLeave"            // hold, end
    };

    private _holdAudio: string[] = [];

    constructor(private conferenceName = "myConference", public name = "action") {

        super(<IActionData> {
            type: "CONFERENCE",
            name: name,
            comments: "",
            params: {
                id               : CoreAction.genId("conference"),
                event            : "join",
                auto_hold        : "false",
                join_muted       : "false",
                mute_on_key      : "off",
                unmute_on_key    : "off",
                exit_on_key      : "off",
                announce         : "none",

                max_participants : "",
                async            : "false",
                $onJoin          : "start",
                $onLeave         : null
            },
            events: {
                [ConferenceOnEvent.OnEnter] : [],
                [ConferenceOnEvent.OnEnterError] : [],
                [ConferenceOnEvent.OnEnterErrorEnded] : [],
                [ConferenceOnEvent.OnEnterErrorMaxParticipants] : [],
                [ConferenceOnEvent.OnEnterAlone] : [],
                [ConferenceOnEvent.OnEnterStarted] : [],
                [ConferenceOnEvent.OnEnterWaiting] : [],
            }
        });
        this.type = Type.Conference;
    }

    public get id (): string {
        return this.getParam("id");
    }

    public set event (evt: ConferenceEvent) {
        this.setParam("event", String(ConferenceEvent[evt]).toLowerCase());
    }
    public get event (): ConferenceEvent {
        switch (this.getParam("event")) {
            case "join":
                return ConferenceEvent.JOIN;
            case "leave":
                return ConferenceEvent.LEAVE;
            case "mute":
                return ConferenceEvent.MUTE;
            case "unmute":
                return ConferenceEvent.UNMUTE;
            case "hold":
                return ConferenceEvent.HOLD;
            case "resume":
                return ConferenceEvent.RESUME;
            default:
                return ConferenceEvent.JOIN;
        }
    }

    public set holdAudio (url: string[]) {
        this._holdAudio = ([] as string[]).concat(url || []);
    }
    public get holdAudio (): string[] {
        return this._holdAudio;
    }

    public set async (val: boolean) {
        this.setParam("async", val ? "true" : "false");
    }
    public get async (): boolean {
        return this.getParam("async") === "true";
    }

    public set announceOnJoin (action: string) {
        this.setParam("announce", action === "beep" ? "beep" : "none");
    }
    public get announeOnJoin (): string {
        return this.getParam("announce") === "beep" ? "beep" : "none";
    }

    public set maxParticipants (val: number) {
        this.setParam("max_participants", typeof val === "number" ? String(val) : "");
    }
    public get maxParticipants (): number {
        return parseInt(this.getParam("max_participants"), 10);
    }

    public set holdOnJoin (val: boolean) {
        if (val) { this.setParam("$onJoin", "hold"); } else if (this.getParam("$onJoin") === "hold") { this.setParam("$onJoin", ""); }
    }
    public get holdOnJoin (): boolean {
        return this.getParam("$onJoin") === "hold";
    }
    public set startOnJoin (val: boolean) {
        if (val) { this.setParam("$onJoin", "start"); } else if (this.getParam("$onJoin") === "start") { this.setParam("$onJoin", ""); }
    }
    public get startOnJoin (): boolean {
        return this.getParam("$onJoin") === "start";
    }

    public set holdOnExit (val: boolean) {
        if (val) { this.setParam("$onLeave", "hold"); } else if (this.getParam("$onLeave") === "hold") { this.setParam("$onLeave", ""); }
    }
    public get holdOnExit (): boolean {
        return this.getParam("$onLeave") === "hold";
    }
    public set endOnExit (val: boolean) {
        if (val) { this.setParam("$onLeave", "end"); } else if (this.getParam("$onLeave") === "end") { this.setParam("$onLeave", ""); }
    }
    public get endOnExit (): boolean {
        return this.getParam("$onLeave") === "end";
    }


    public set autoHold (val: boolean) {
        this.setParam("auto_hold", val ? "true" : "false");
    }
    public get autoHold (): boolean {
        return this.getParam("auto_hold") === "true";
    }

    public set joinMuted (val: boolean) {
        this.setParam("join_muted", val ? "true" : "false");
    }
    public get joinMuted (): boolean {
        return this.getParam("join_muted") === "true";
    }

    public set muteOnKey (key: string) {
        this.setParam("mute_on_key", String(key));
    }
    public get muteOnKey (): string {
        return this.getParam("mute_on_key");
    }

    public set unmuteOnKey (key: string) {
        this.setParam("unmute_on_key", String(key));
    }
    public get unmuteOnKey (): string {
        return this.getParam("unmute_on_key");
    }

    public set exitOnKey (key: string) {
        this.setParam("exit_on_key", String(key));
    }
    public get exitOnKey (): string {
        return this.getParam("exit_on_key");
    }

    public toCCML (): string {

        let id = CoreAction.genId();

        let ccml = "";

        let p = VarParser.parseVars(this.conferenceName);
        this.conferenceName = p[0];
        ccml += p[1].join("");

        // Build OnEnter Event Handlers
        if (
            this.getParam("event") === "join" &&
            this.events[ConferenceOnEvent.OnEnterAlone].length +
            this.events[ConferenceOnEvent.OnEnterStarted].length +
            this.events[ConferenceOnEvent.OnEnterWaiting].length +
            this.events[ConferenceOnEvent.OnEnter].length > 0
        ) {
            let params = super.toCCML({
                    event            : "control",
                    auto_hold        : "holdAuto",
                    mute_on_key      : "mute",
                    unmute_on_key    : "unmute",
                    exit_on_key      : "leave",
                    max_participants : "maxParticipants",
                    $onJoin          : "onJoin",            // hold, start
                });

            ccml += `
                <CONFERENCE errorActionId="${id}_errorHandler" ${params} joinMuted="true" async="true">${this.escapeXml(this.conferenceName)}</CONFERENCE>
                <CONDITION>
                    <WAIT conferenceName="${this.escapeDoubleQuotes(this.conferenceName)}" conferenceState="START" nextActionId="${id}_onEnterStarted"/>
                    <WAIT conferenceName="${this.escapeDoubleQuotes(this.conferenceName)}" conferenceState="END" nextActionId="${id}_onEnterErrorEnded"/>
                    <WAIT condition="true" nextActionId="${id}_onEnterWaiting"/>
                </CONDITION>
                <BOOKMARK id="${id}_onEnterStarted"/>
                    ${this.events[ConferenceOnEvent.OnEnterStarted].map(e => e.toCCML()).join("")}
                    <WAIT condition="true" nextActionId="${id}_onEnter"/>
                <BOOKMARK id="${id}_onEnterWaiting"/>
                    ${this.events[ConferenceOnEvent.OnEnterWaiting].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_onEnter"/>
                    <CONDITION>
                        <WAIT condition="! \${conference.participants} || \${conference.participants} &lt; 2" nextActionId="${id}_onEnterAlone"/>
                        <WAIT condition="true" nextActionId="${id}_onEnterDefault"/>
                    </CONDITION>
                    <BOOKMARK id="${id}_onEnterAlone"/>
                        ${this.events[ConferenceOnEvent.OnEnterAlone].map(e => e.toCCML()).join("")}
                    <BOOKMARK id="${id}_onEnterDefault"/>
                        ${this.events[ConferenceOnEvent.OnEnter].map(e => e.toCCML()).join("")}
            `;
            // Seems to have been fixed.
            // <CONFERENCE control="leave">${this.escapeXml(this.conferenceName)}</CONFERENCE>
            // CONFERENCE control=leave because joining the same conference twice throws an error.
        }

        // Build OnError Event Handlers.
        if (this.getParam("event") === "join" && this.events[ConferenceOnEvent.OnEnterErrorMaxParticipants].length + this.events[ConferenceOnEvent.OnEnterErrorEnded].length + this.events[ConferenceOnEvent.OnEnterError].length > 0) {

            let params = super.toCCML({
                    id               : "id",
                    event            : "control",
                    auto_hold        : "holdAuto",
                    join_muted       : "joinMuted",
                    mute_on_key      : "mute",
                    unmute_on_key    : "unmute",
                    exit_on_key      : "leave",
                    announce         : "announce",
                    max_participants : "maxParticipants",
                    async            : "async",             // boolean
                    $onJoin          : "onJoin",            // hold, start
                    $onLeave         : "onLeave"            // hold, end
                });

            ccml += `<CONFERENCE errorActionId="${id}_errorHandler" ${params}>${this.escapeXml(this.conferenceName)}`;
            this.holdAudio.forEach(url => ccml += `<HOLDMUSIC url="${url}"/>`);
            ccml += `</CONFERENCE>`;

            ccml += `
                <WAIT condition="true" nextActionId="${id}_done"/>
                <CONDITION id="${id}_errorHandler">
                    <WAIT error="confFull" nextActionId="${id}_onEnterErrorMaxParticipants"/>
                    <WAIT error="confEnded" nextActionId="${id}_onEnterErrorEnded"/>
                    <WAIT condition="true" nextActionId="${id}_onEnterError"/>
                </CONDITION>

                <BOOKMARK id="${id}_onEnterErrorMaxParticipants"/>
                    ${this.events[ConferenceOnEvent.OnEnterErrorMaxParticipants].map(e => e.toCCML()).join("")}
                    <WAIT condition="true" nextActionId="${id}_onEnterError"/>
                <BOOKMARK id="${id}_onEnterErrorEnded"/>
                    ${this.events[ConferenceOnEvent.OnEnterErrorEnded].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_onEnterError"/>
                    ${this.events[ConferenceOnEvent.OnEnterError].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_done"/>
            `;
        } else {

            ccml += `<CONFERENCE errorActionId="${id}_errorHandler" ${super.toCCML(this._paramMap)}>${this.escapeXml(this.conferenceName)}`;
            this.holdAudio.forEach(url => ccml += `<HOLDMUSIC url="${url}"/>`);
            ccml += `</CONFERENCE>
                <BOOKMARK id="${id}_onEnterErrorEnded"/><BOOKMARK id="${id}_errorHandler"/>
            `;
        }

        return ccml;
    }
}
