import {
    IActionData,
    CoreAction,
    Type,
} from "./interfaces";
import { Say } from "./Say";


export const DetectOnEvent: any = {
    OnLivePerson     : "on_live_person",
    OnVoicemailEarly : "on_voicemail_early",
    OnVoicemail      : "on_voicemail"
};

/**
     {
         "type": "DETECT",

        "params": {
            "name": "inbound-detect-voicemail",
            "triggerLiveInMS": "3000",
        },

        "events": {
            "on_live_person": [],
            "on_voicemail_early": [],
            "on_voicemail": []
        }
     }
    */
export class Detect extends CoreAction {

    // private _paramMap: any = {
    //     id: "id",
    // };
    private _triggerLiveInMS = 3000;

    constructor(public name = "action") {

        super(<IActionData> {
            type: "DETECT",
            name: name,
            comments: "",
            params: {
                id : CoreAction.genId("detect"),
            },
            events: {
                [DetectOnEvent.OnLivePerson] : [],
                [DetectOnEvent.OnVoicemailEarly] : [],
                [DetectOnEvent.OnVoicemail] : [],
            }
        });
        this.type = Type.Detect;
    }

    public get id (): string {
        return this.getParam("id");
    }

    public get triggerLiveInMS () {
        return this._triggerLiveInMS;
    }

    public set triggerLiveInMS (v: number) {
        this._triggerLiveInMS = Math.max(+v || 3000, 100);
    }


    public toCCML (): string {

        // If no onVoiceMailEarly events, play the BRRR.
        const playBrrr = `
            <STOPAUDIO />
            <PARALLEL scriptId="___detect_playBrrr___" action="start" />
            <WAIT id="___detect_silence___" silence="1" />
            <CONDITION>
                <WAIT noise="1" nextActionId="___detect_silence___" />
                <TIMER milliSeconds="1600" waitForIdle="false" nextActionId="` + "${next}" + `" />
            </CONDITION>
            <PARALLEL scriptId="___detect_playBrrr___" action="stop" />
        `;
        // Embed this in SAY/PLAYs in the onLivePerson event.
        const liveDetection = `
            <CONDITION>
              <WAIT callProgress="AM" nextActionId="___detect_endStateRouter___">
                <STOPAUDIO />
              </WAIT>
              <WAIT callState="CONNECT" nextActionId="___detect_endStateRouter___" />
              <TIMER seconds="300" nextActionId="___detect_endStateRouter___" />
            </CONDITION>
        `;
        // Embed this in SAY/PLAYs in the onVoicemail events.
        const vmDetection = `
            <CONDITION>
              <WAIT noise="500" nextActionId="___detect_onVoicemailEarly___">
                <STOPAUDIO />
              </WAIT>
              <TIMER milliSeconds="2000" nextActionId="` + "${next}" + `"/>
            </CONDITION>
        `;

        // OnLivePerson required. Play error message if empty.
        if (this.events[DetectOnEvent.OnLivePerson].length === 0) {
            this.events[DetectOnEvent.OnLivePerson] = [new Say("Error. The detect on live person event requires an action.", "onLivePersonError")];
        }
        // OnVoicemail required. Play error message if empty.
        if (this.events[DetectOnEvent.OnVoicemail].length === 0) {
            this.events[DetectOnEvent.OnVoicemail] = [new Say("Error. The detect on voice mail event requires an action.", "onVoiceMailError")];
        } else if (this.events[DetectOnEvent.OnVoicemailEarly].length === 0 && this.events[DetectOnEvent.OnVoicemail][0].type === Type.HangUp) {
        // If no OnVoicemailEarly and first OnVoiceMail action is HANGUP, do hangup early.
            this.events[DetectOnEvent.OnVoicemailEarly] = [this.events[DetectOnEvent.OnVoicemail][0]];
            this.events[DetectOnEvent.OnVoicemail] = [];
        }

        this.events[DetectOnEvent.OnLivePerson].filter(e => e.type === Type.Say || e.type === Type.Play).forEach((e: Say) => e.injectCCML = liveDetection);
        this.events[DetectOnEvent.OnVoicemailEarly].filter(e => e.type === Type.Say || e.type === Type.Play).forEach((e: Say) => e.injectCCML = vmDetection);
        this.events[DetectOnEvent.OnVoicemail].filter(e => e.type === Type.Say || e.type === Type.Play).forEach((e: Say) => e.injectCCML = vmDetection);

        const ccml = `
            <DSP  varName="dspA" detectTones="737, 738, 850, 870, 1000, 1005"
                  silenceDeadBand="600"
                  voiceDeadBand="200"
                  voiceThreshold="-45dB"
            />

            <WAIT callState="CONNECT" />

            <CONDITION>
                <WAIT callProgress="VOICE" nextActionId="___detect_onLivePerson___"/>
                <WAIT callProgress="AM" nextActionId="___detect_endStateRouter___"/>
                <WAIT callProgress="any" nextActionId="___detect_endStateRouter___" />
                <TIMER milliSeconds="${ this.triggerLiveInMS }" waitForIdle="false" nextActionId="___detect_onLivePerson___" />
            </CONDITION>

            <BOOKMARK id="___detect_endStateRouter___" />
            <CONDITION>
                <WAIT callProgress="AM" nextActionId="___detect_onVoicemailEarly___" />
                <WAIT callProgress="VOICE" nextActionId="___detect_done___" />
                <WAIT condition="true" nextActionId="___detect_defaultError___" />
            </CONDITION>

            <BOOKMARK id="___detect_defaultError___"/>
            <HANGUP />

            <SUBSCRIPT id="___detect_playBrrr___" inheritGlobalConditions="false">
                <TIMER milliSeconds="1000" />
                <PLAYTONE id="___detect_brrrTone___" freq1="75" freq2="10" ampl1="-15" ampl2="-15" duration="2000" />
                <TIMER milliSeconds="3000" nextActionId="___detect_brrrTone___" />
            </SUBSCRIPT>

            <CONDITION global="true">
                <WAIT tone="any" toneDuration="400">
                  <CONDITION>
                    <WAIT condition="` + "${connectTime}" + ` &gt; 2" nextActionId="___detect_onVoicemailEarly___">
                      <STOPAUDIO />
                    </WAIT>
                    <WAIT condition="true" nextActionId="` + "${next}" + `" />
                  </CONDITION>
                </WAIT>
                <TIMER seconds="300" nextActionId="___detect_defaultError___" />
            </CONDITION>

            <BOOKMARK id="___detect_onLivePerson___"/>
                ${ this.events[DetectOnEvent.OnLivePerson].map(e => e.toCCML()).join("") }
            <WAIT nextActionId="___detect_done___" />

            <BOOKMARK id="___detect_onVoicemailEarly___"/>
                ${ this.events[DetectOnEvent.OnVoicemailEarly].length === 0 ? playBrrr : this.events[DetectOnEvent.OnVoicemailEarly].map(e => e.toCCML()).join("") }

            <BOOKMARK id="___detect_onVoicemail___"/>
                ${ this.events[DetectOnEvent.OnVoicemail].map(e => e.toCCML()).join("") }

            <BOOKMARK id="___detect_done___" />
            <WAIT callState="CONNECT" />
            <DSP action="stop" />
        `;

        return ccml;
    }
}
