import {
    IActionData,
    CoreAction,
    Type,
} from "./interfaces";
import { MD5 } from "../utils";

/**
 * https://developers.google.com/analytics/devguides/collection/protocol/v1/parameters
 */
export class GoogleAnalytics extends CoreAction {

    private _paramMap: any = {
        id      : "id",
        url     : "url",
        method  : "method",
        timeout : "timeout",
        async   : "async",
    };

    private _gaParams: any = {
        "tid": "${___config___.google_analytics.tracking_id}",
        "an": "${___config___.google_analytics.source}",
        "v": "1",
        "aip": "1",
        "ds": "phone",
        "geoid": "US",
        "av": "1.0",
    };

    /**
     *
     * {
     *  "type": "GOOGLEANALYTICS",
     *  "callback_response": { app_id: string; api_no: string; call_id: string; caller_no: string; },
     *  "event": { action: string, label = '', category = 'action', value = 1 }
     *  "params": { Google Parameters }
     * }
     *
     * @param params
     * @param callbackResponse
     * @param name
     */
    constructor(callbackResponse?: { app_id: string; api_no: string; call_id: string; caller_no: string; }, params?: { [key: string]: string; }, public name = "action") {

        super(<IActionData> {
            type: "GOOGLEANALYTICS",
            name: name,
            comments: "",
            params: {
                id      : CoreAction.genId("google_analytics"),
                varName : "___google-analytics_var___",
                url     : "https://ssl.google-analytics.com/collect",
                method  : "post",
                timeout : "30",
                async   : "true",
            }}
        );
        this.type = Type.GoogleAnalytics;

        this._gaParams.dt = name || undefined;
        if (callbackResponse !== undefined) {
            this._gaParams.aid = callbackResponse.app_id;
            this._gaParams.uid = MD5.hash(callbackResponse.caller_no);
            this._gaParams.cid = MD5.hash(`${callbackResponse.call_id}-${callbackResponse.app_id}`);
            this._gaParams.cd = callbackResponse.api_no + "/" + name;
            this._gaParams.dh = callbackResponse.api_no;
            this._gaParams.dp = callbackResponse.api_no + "/" + name;
        }

        if (params !== undefined) {
            for (let k in params) {
                if (params.hasOwnProperty(k)) {
                    this._gaParams[k] = String(params[k]);
                }
            }
        }
    }

    get id (): string {
        return this.getParam("id");
    }

    public setGAEvent (action: string, label = "", category = "action", value = 1) {
        this._gaParams.t = "event";
        this._gaParams.ec = "action";
        this._gaParams.ea = action;
        this._gaParams.el = label || undefined;
        this._gaParams.ec = category;
        this._gaParams.ev = String(value);
    }
    public getGAEvent () {
        return {
            action: this._gaParams.ea,
            label: this._gaParams.el,
            category: this._gaParams.ec,
            value: +this._gaParams.ev
        };
    }

    public toCCML (): string {

        let ccml = `<REQUEST ${super.toCCML(this._paramMap)}>`;

        for (let k in this._gaParams) {
            if (this._gaParams.hasOwnProperty(k)) {
                ccml += `<SET varName="${k}" value=${JSON.stringify(this._gaParams[k])} export="false"/>`;
            }
        }
        ccml += "</REQUEST>";

        return ccml.replace(/[ \n]+/g, " ").trim();
    }
}

/*
 https://ssl.google-analytics.com/collect

 {
     "tracking_id": "",		// tid=UA-122323
     "source": "",			// an=source

     "uid": "md5(caller_no)",
     "cid": "md5(call_id +'-'+ app_id)",

     "version": "1",			// v=1
     "aip": "1",
     "ds": "phone",
     "geoid": "US",
     "av": "1.0",
     "cd": "api_no/name",
     "dh": "api_no",
     "dt": "comment",
     "dp": "api_no/name",


     "t": "event",
     "ec": "action",
     "ea": "TYPE",
     "el": "name || ''",
     "ev": "1"
 }

 <REQUEST id="" url="https://ssl.google-analytics.com/collect" method="POST" async="true">

 <SET varName="tid" value="${___config___.google_analytics.tracking_id}" export="false"/>   // Tracking ID
 <SET varName="an" value="${___config___.google_analytics.source}" export="false"/>			// Source

 <SET varName="t" value="event|screenview" export="false"/>			// Hit Type
 <SET varName="cd" value="" export="false"/>							// Screen Name

 <SET varName="ec" value="" export="false"/>							// Event Catagory
 <SET varName="ea" value="" export="false"/>							// Event Action
 <SET varName="el" value="" export="false"/>							// Event Label
 <SET varName="ev" value="1" export="false"/>						// Event Value

 <SET varName="dh" value="${api_no}" export="false"/>				// Doc Hostname
 <SET varName="dt" value="${comment}" export="false"/>				// Doc Title
 <SET varName="dp" value="${api_no/name" export="false"/>			// Doc Path

 <SET varName="aid" value="${app_id}" export="false"/>
 <SET varName="uid" value="$MD5{caller_no}" export="false"/>
 <SET varName="cid" value="$MD5{call_id-app_id}" export="false"/>

 <SET varName="v" value="1" export="false"/>
 <SET varName="aip" value="1" export="false"/>
 <SET varName="ds" value="phone" export="false"/>
 <SET varName="geoid" value="US" export="false"/>
 <SET varName="av" value="1.0" export="false"/>

 </REQUEST>
 */
