import {
    IActionData,
    CoreAction,
    Type,
} from "./interfaces";
import { VarParser } from "../utils";

export enum QueueEvent {
    OFFER, POLL, POLL_CONNECT, REMOVE
}

export const QueueOnEvent: any = {
    OnEnter              : "on_enter",
    OnEnterError         : "on_enter_error",
    OnEnterErrorMaxCalls : "on_enter_error_max_calls"
};

/**
     {
         "type": "CONFERENCE",

        "params": {
            "name": "my-conference",
            "event": "offer|poll|poll_connect|remove",
            "async": false,
            "max_calls": 20,
            "hold_audio": ["http://file.wav"],
        },

        "events": {
            "on_enter": [],

            "on_enter_error": [],
            "on_enter_error_max_calls": []
        }
        }
    */
export class Queue extends CoreAction {

    private _paramMap: any = {
        id        : "id",
        event     : "control",
        max_calls : "maxCalls",
        async     : "async",     // boolean
    };

    private _holdAudio: string[] = [""]; // Empty string means play default hold music.
    private _event: QueueEvent = QueueEvent.OFFER;

    constructor(private queueName = "myQueue", public name = "action") {

        super(<IActionData> {
            type: "QUEUE",
            name: name,
            comments: "",
            params: {
                id        : CoreAction.genId("queue"),
                event     : "enqueue",
                max_calls : null,
                async     : "false",
            },
            events: {
                [QueueOnEvent.OnEnter] : [],
                [QueueOnEvent.OnEnterError] : [],
                [QueueOnEvent.OnEnterErrorMaxCalls] : [],
            }
        });
        this.type = Type.Queue;
    }

    public get id (): string {
        return this.getParam("id");
    }

    public set event (evt: QueueEvent) {
        this._event = evt;
        this.setParam("event", evt === (QueueEvent.POLL || QueueEvent.POLL_CONNECT) ? "dequeue" : evt === QueueEvent.REMOVE ? "removeMe" : "enqueue");
        if (this.getParam("event") !== "enqueue") {
            this.setParam("async", "true");
        }
    }
    public get event (): QueueEvent {
        return this._event;
    }

    public set holdAudio (url: string[] | boolean) {
        if (url === true) {
            this._holdAudio = [""]; // Plays default
        } else if (url === false) {
            this._holdAudio = [];   // Plays none
        } else {
            this._holdAudio = ([] as string[]).concat(url);
        }
    }
    public get holdAudio (): string[] | boolean {
        if (this._holdAudio.length === 0) {
            return false;
        } else if (this._holdAudio[0] === "") {
            return true;
        }
        return this._holdAudio.slice(0);
    }

    public set async (val: boolean) {
        this.setParam("async", val ? "true" : "false");
        if (this.getParam("event") !== "enqueue") {
            this.setParam("async", "true");
        }
    }
    public get async (): boolean {
        return this.getParam("async") === "true";
    }

    public set maxCalls (val: number | undefined) {
        this.setParam("max_calls", typeof val === "number" ? String(val) : "");
    }
    public get maxCalls (): number | undefined {
        return this.getParam("max_calls") ? parseInt(this.getParam("max_calls"), 10) : undefined;
    }


    public toCCML (): string {

        let id = CoreAction.genId();
        let ccml = "";

        let p = VarParser.parseVars(this.queueName);
        this.queueName = p[0];
        ccml += p[1].join("");

        // Build OnEnter Event Handlers
        if (this._event === QueueEvent.OFFER && this.events[QueueOnEvent.OnEnter].length > 0) {
            // Make unique ID, Force async=true
            let params = super.toCCML(this._paramMap)
                .replace("id=\"", "id = \"2_")
                .replace("async=\"false\"", "async=\"true\"");
            ccml += `<CALLQUEUE errorActionId="${id}_errorHandler"${params}><![CDATA[${this.queueName}]]></CALLQUEUE>
                    ${this.events[QueueOnEvent.OnEnter].map(e => e.toCCML()).join("")}`;
        }

        // Build OnError Event Handlers.
        if (this._event === QueueEvent.OFFER && (this.events[QueueOnEvent.OnEnterErrorMaxCalls].length > 0 || this.events[QueueOnEvent.OnEnterError].length > 0)) {

            ccml += `<CALLQUEUE errorActionId="${id}_errorHandler"${super.toCCML(this._paramMap)}><![CDATA[${this.queueName}]]>`;
            this._holdAudio.forEach(url => ccml += url !== "" ? `<HOLDMUSIC url="${url}"/>` : "<HOLDMUSIC/>");
            ccml += `</CALLQUEUE>`;

            ccml += `
                <WAIT condition="true" nextActionId="${id}_done"/>
                <CONDITION id="${id}_errorHandler">
                    <WAIT error="queueFull" nextActionId="${id}_onEnterErrorMaxCalls"/>
                    <WAIT error="any" nextActionId="${id}_onEnterError"/>
                </CONDITION>
                <BOOKMARK id="${id}_onEnterErrorMaxCalls"/>
                    ${this.events[QueueOnEvent.OnEnterErrorMaxCalls].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_onEnterError"/>
                    ${this.events[QueueOnEvent.OnEnterError].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_done"/>
            `;

        } else if (this._event === QueueEvent.POLL_CONNECT) {

            ccml += `
                <CALLQUEUE errorActionId="${id}_onEnterError" control="dequeue" async="true"><![CDATA[${this.queueName}]]></CALLQUEUE>
                <CONDITION>
                    <WAIT condition="\${queueEntry} !== null" />
                    <WAIT condition="true" nextActionId="${id}_onEnterError" />
                </CONDITION>
                <SET varName="___lastQueuePollConnectCall___" export="true" evaluate="true"><![CDATA[ JSON.stringify({id:\${queueEntry.callId}, no:\${queueEntry.phoneNo}}) ]]></SET>
                <CALLTRANSFER errorActionId="${id}_onEnterError" async="true">\${queueEntry}</CALLTRANSFER>
                <WAIT callTransfer="routed"/>
                ${this.events[QueueOnEvent.OnEnter].map(e => e.toCCML()).join("")}
                <CONDITION>
                    <WAIT callTransfer="ended" nextActionId="${id}_done"/>
                    <TIMER seconds="14400" nextActionId="${id}_done" />
                </CONDITION>
                <WAIT nextActionId="${id}_done"/>
                <BOOKMARK id="${id}_onEnterError"/>
                    ${this.events[QueueOnEvent.OnEnterError].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_done" />
            `;

        } else if (this._event === QueueEvent.POLL) {

            ccml += `
                <CALLQUEUE ${super.toCCML(this._paramMap)}><![CDATA[${this.queueName}]]></CALLQUEUE>
                <CONDITION>
                    <WAIT condition="\${queueEntry} !== null" />
                    <WAIT condition="true" nextActionId="${id}_errorHandler" />
                </CONDITION>
                <SET varName="___lastQueuePollConnectCall___" export="true" evaluate="true"><![CDATA[ JSON.stringify({id:\${queueEntry.callId}, no:\${queueEntry.phoneNo}}) ]]></SET>
                ${this.events[QueueOnEvent.OnEnter].map(e => e.toCCML()).join("")}
                <WAIT nextActionId="${id}_done" />
                <BOOKMARK id="${id}_errorHandler"/>
                    ${this.events[QueueOnEvent.OnEnterError].map(e => e.toCCML()).join("")}
                <BOOKMARK id="${id}_done" />
            `;

        } else {

            ccml += `<CALLQUEUE ${super.toCCML(this._paramMap)}><![CDATA[${this.queueName}]]>`;
            if (this._event === QueueEvent.OFFER) {
                this._holdAudio.forEach(url => ccml += url !== "" ? `<HOLDMUSIC url="${url}"/>` : "<HOLDMUSIC/>");
            }
            ccml += `</CALLQUEUE><BOOKMARK id="${id}_errorHandler"/>`;
        }
        // For poll_connect, mostly to be safe if <HANGUP> is right after.
        if (this._event === QueueEvent.OFFER && this.async === false) {
            ccml += "<TIMER milliSeconds=\"1000\" waitForIdle=\"false\"/>";
        }

        return ccml;
    }
}
