import {
    IAction,
    IActionData,
    CoreAction,
    Type,
} from "./interfaces";
import { VarParser } from "../utils";

export const WebhookOnEvent: any = {
    OnError: "on_error",
};


export class Webhook extends CoreAction {

    private _paramMap: any = {
        id                  : "id",
        url                 : "url",
        url2                : "url2",
        url3                : "url3",
        method              : "method",
        failureRetryMax     : "failureRetryMax",
        failureRetryDelay   : "failureRetryDelay",
        timeout             : "timeout",
        async               : "async",
        varName             : "varName",
        includeExportedVars : "includeExportedVars",
    };
    private _url: string;
    private _method: string;
    private _headers: { [key: string]: string };
    private _contentType: string;
    private _webhookData = "{}";
    private _webhookDataType = "object";
    private _awaitResponse = false;


    constructor(public eventName = "webhook", public name = "action") {

        super(<IActionData> {
            type: "WEBHOOK",
            name: name,
            comments: "",
            params: {
                id                  : CoreAction.genId("webhook"),
                varName             : "___webhookDataResponse___",
                url                 : "${___internalCallbackUrl___}",
                url2                : "${___internalCallbackUrl2___}",
                url3                : "${___internalCallbackUrl3___}",
                method              : "post",
                failureRetryMax     : "0",
                failureRetryDelay   : "100",
                timeout             : "30",
                async               : "true",
                includeExportedVars : "true",
            },
            events: {
                [WebhookOnEvent.OnError]: [],
            }
        });
        this.type = Type.Webhook;
    }

    public get id (): string {
        return this.getParam("id");
    }

    public set url (url: string) {
        this._url = String(url);
    }
    public get url (): string {
        return this._url;
    }

    public set method (method: string) {
        this._method = String(method);
    }
    public get method (): string {
        return this._method;
    }

    public set headers (headers: { [key: string]: string }) {
        this._headers = headers;
    }
    public get headers (): { [key: string]: string } {
        return this._headers;
    }

    public set contentType (contentType: string) {
        this._contentType = String(contentType);
    }
    public get contentType (): string {
        return this._contentType;
    }

    public set webhookData (data: any) {
        this._webhookDataType = typeof data;
        this._webhookData = JSON.stringify(data);
    }
    public get webhookData (): any {
        return JSON.parse(this._webhookData);
    }

    public set awaitResponse (v: boolean) {
        this._awaitResponse = v;
    }
    public get awaitResponse () {
        return this._awaitResponse;
    }

    public toCCML (): string {

        const id = CoreAction.genId();

        // Parse any custom data sent by the Webhook.
        const dataSet = (function (type: string, data: string) {

            if (type !== "string") {
                return `<SET varName="webhook_data" export="false"><![CDATA[${data}]]></SET>`;
            }
            try {
                const exp = VarParser.parseExpression(JSON.parse(data));
                return `<SET varName="webhook_data" evaluate="true" export="false"><![CDATA[ JSON.stringify( ${exp} ) ]]></SET>`;
            } catch (err) {
                return `<SET varName="webhook_data" export="false"><![CDATA[${data}]]></SET>`;
            }

        }(this._webhookDataType, this._webhookData));

        // Handle events
        const eventHandler = (function (isAwait: boolean, onError: IAction[]) {

            if (isAwait) {

                return `
                    <WAIT nextActionId="${id}_onSuccess"/>
                    <BOOKMARK id="${id}_onError"/>
                        <SET varName="___webhookData___" value="\${NULL}" export="false"/>
                        ${onError.map(e => e.toCCML()).join("")}
                        <WAIT nextActionId="${id}_onSkipSuccess"/>
                    <BOOKMARK id="${id}_onSuccess"/>
                    <SET varName="___webhookData___" value="\${___webhookDataResponse___.data}" export="false"/>
                    <BOOKMARK id="${id}_onSkipSuccess"/>
                `;

            } else {

                return `
                    <WAIT nextActionId="${id}_onSuccess"/>
                    <BOOKMARK id="${id}_onError"/>
                        ${onError.map(e => e.toCCML()).join("")}
                    <BOOKMARK id="${id}_onSuccess"/>
                `;
            }

        }(this._awaitResponse, this.events[WebhookOnEvent.OnError]));

        if (this._awaitResponse || this.events[WebhookOnEvent.OnError].length > 0) {
            this.setParam("async", "false");
        }

        return `
            <REQUEST ${super.toCCML(this._paramMap)} errorActionId="${id}_onError">
                ${dataSet}
                <SET varName="___is_webhook_action___" value="true" export="false"/>
                <SET varName="___is_webhook_await___" value="${this._awaitResponse.toString()}" export="false"/>
                <SET varName="webhook_event_name" value="${this.eventName}" export="false"/>
                <SET varName="config_callback_url" value="${this._url || "${___config___.callback.url}"}"/>
                <SET varName="config_callback_method" value="${this._method || "${___config___.callback.method}"}"/>
                <SET varName="config_callback_headers"><![CDATA[ ${this.headers ? JSON.stringify(this.headers) : "${___config___.callback.headers}"} ]]></SET>
                <SET varName="config_callback_content_type" value="${this._contentType || "${___config___.callback.content_type}"}"/>
                <SET varName="callTime" value="\${callTimeMS}"/>
                <SET varName="last_queue_poll_connect_call" value="\${___lastQueuePollConnectCall___}"/>
                <SET varName="session_data" value="\${___sessionData___}"/>
            </REQUEST>
            ${eventHandler}
        `;
    }
}
